/***********************************************************
* Copyright (c) 2009-2011 OpenVox Communication Co.,Ltd.
* All Rights Reserved.	
* Author: mark.liu@openvox.cn
***********************************************************/
/*
 * Filename:
		epcs.c
 * Description:
 		A flash update programme for OpenVox G400P and A2410P
 * Version:
 		1.0
 * Created:
 		02/08/2010
 * Author:
 		Written by mark.liu
 * Company: OpenVox
 *
 */

#include <stdio.h>
#include "lib/pci.h"

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <syslog.h>
#include <fcntl.h>
#include <pthread.h>

//#include <strings.h> /* strncasecmp */
#include "epcs.h"
#include "cards.h"

#include "menu.h"
#include "altera_avalon_spi_regs.h"

#include <signal.h>

//#define DEBUG

/************************************************************************
 * global variants
 ************************************************************************/
static int max_epcs_dev;                    // PCI card number 忨
static t_opvx_pci_dev_s *popvx_pci_dev;     // epcs number
static t_menu_item *pmenu_epcs_flash;	    // menu itmes
static char *version = "v1.3.5 2014-03-17"; // version
static int card_numbers[OPVX_END];          // ÿһ͵Ŀ
static int card_indexes[EPCS_MAX_DEVICE];   // ÿһӦĿ
static int card_menu_indexes[OPVX_END];     // ÿһ͵ĿӦ
static int card_type_number;                // м͵Ŀ

pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

/************************************************************************
 * menu items
 ************************************************************************/
static t_menu_item menu_main[] = 
{
	{MENU_CARD_UPDATE,	"1. OpenVox EPCS Flash Update Utility",	"Update", 0},
	{MENU_CARD_BACKUP,	"2. OpenVox EPCS Flash Backup Utility",	"Backup", 0},
	{MENU_CARD_EDIT,	"3. OpenVox EPCS Flash Edit   Utility",	"Edit", 0},
};


//altera_avalon_spi.c
// This is a very simple routine which performs one SPI master transaction. 
// It would be possible to implement a more efficient version using interrupts and sleeping threads but this is probably not worthwhile initially.
static int alt_avalon_spi_command(alt_u8 *u_base, alt_u32 slave, alt_u32 write_length, const alt_u8 *write_data, alt_u32 read_length, alt_u8 *read_data, alt_u32 flags)
{
    const alt_u8 *write_end = write_data + write_length;
    alt_u8 *read_end = read_data + read_length;

    alt_u32 write_zeros = read_length;
    alt_u32 read_ignore = write_length;
    alt_u32 status;
    //volatile alt_u8 status;

    // We must not send more than two bytes to the target before it has returned any as otherwise it will overflow. 
    // Unfortunately the hardware does not seem to work with credits > 1, leave it at 1 for now.
    alt_32 credits = 1;

    alt_u8 discard;

	alt_u16 delay=0xff;
	//-alt_u8 *base = (alt_u8 *)(u_base);
	alt_u8 *base = (u_base);

	if(pthread_mutex_lock(&mutex) != 0) {
		fprintf(stderr, "Lock failed.\n");
		return -1;
	}

    // Warning: this function is not currently safe if called in a multi-threaded environment, something above must perform locking to make it safe if more than one thread intends to use it.

    //IOWR_ALTERA_AVALON_SPI_SLAVE_SEL(base, 1 << slave);
    //slave=0
    base[ALTERA_AVALON_SPI_SLAVE_SEL_REG] = (1 << slave);//&0xff;
    //printf("epcs4_reg[%x]=%x\n", baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG, epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG]);
    //Щλֵܸ
    //epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 1] = ((1 << slave)>>8) & 0xff;
    //epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 2] = ((1 << slave)>>16)& 0xff;
    //epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 3] = ((1 << slave)>>24)& 0xff;
    //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 1]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 1]);
    //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 2]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 2]);
    //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 3]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_SLAVE_SEL_REG + 3]);

    // Set the SSO bit (force chipselect) only if the toggle flag is not set
    if ((flags &ALT_AVALON_SPI_COMMAND_TOGGLE_SS_N) == 0)
    {
        //IOWR_ALTERA_AVALON_SPI_CONTROL(base, ALTERA_AVALON_SPI_CONTROL_SSO_MSK);
        //ֽֿڣֱֵ
        base[ALTERA_AVALON_SPI_CONTROL_REG] = ALTERA_AVALON_SPI_CONTROL_SSO_MSK & 0xff;
        base[ALTERA_AVALON_SPI_CONTROL_REG + 1] = ALTERA_AVALON_SPI_CONTROL_SSO_MSK >> 8;
        //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG]);
        //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 1]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 1]);
        //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 2]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 2]);
        //printf("epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 3]=%x\n", epcs4_reg[baseoff + ALTERA_AVALON_SPI_CONTROL_REG + 3]);
    }

    // Discard any stale data present in the RXDATA register, in case previous communication was interrupted and stale data was left behind.
    //IORD_ALTERA_AVALON_SPI_RXDATA(base);
    discard = base[ALTERA_AVALON_SPI_RXDATA_REG];
    //discard = epcs4_reg[baseoff + ALTERA_AVALON_SPI_RXDATA_REG + 1];
    //discard = epcs4_reg[baseoff + ALTERA_AVALON_SPI_RXDATA_REG + 2];
    //discard = epcs4_reg[baseoff + ALTERA_AVALON_SPI_RXDATA_REG + 3];
#if 0
        //ȴֱ׼á׼
        do
        {
            //status = IORD_ALTERA_AVALON_SPI_STATUS(base);
            status = base[ALTERA_AVALON_SPI_STATUS_REG];
		//wujc -s
		while(delay)
			delay--;
		//wujc -e
            printf(" rx/tx status=0x%x\n", status);
            //printf("%c", 0x8);
        }
        while (((status &ALTERA_AVALON_SPI_STATUS_TRDY_MSK) == 0 || credits == 0) && (status &ALTERA_AVALON_SPI_STATUS_RRDY_MSK) == 0);
#endif
    // Keep clocking until all the data has been processed.
    for (;;)
    {
#if 0
        //ȴֱ׼á׼
        do
        {
            //status = IORD_ALTERA_AVALON_SPI_STATUS(base);
            status = base[ALTERA_AVALON_SPI_STATUS_REG];
		//wujc -s
		while(delay)
			delay--;
		//wujc -e
            printf("status=0x%x\n", status);
            //printf("%c", 0x8);
        }
        while (((status &ALTERA_AVALON_SPI_STATUS_TRDY_MSK) == 0 || credits == 0) && (status &ALTERA_AVALON_SPI_STATUS_RRDY_MSK) == 0);
#endif
        status = base[ALTERA_AVALON_SPI_STATUS_REG];
        if ((status &ALTERA_AVALON_SPI_STATUS_TRDY_MSK) != 0 && credits > 0)
        {
            credits--;

            //ݣͷ
            if (write_data < write_end)
            {
                //IOWR_ALTERA_AVALON_SPI_TXDATA(base,  *write_data++);
                //printf("%x ", *write_data);
                base[ALTERA_AVALON_SPI_TXDATA_REG] = *write_data++;
            }
            //ûݷͶҪգͷ0
            else if (write_zeros > 0)
            {
                write_zeros--;
                //IOWR_ALTERA_AVALON_SPI_TXDATA(base, 0);
                base[ALTERA_AVALON_SPI_TXDATA_REG] = 0;
            }
            else
            {
                credits =  - 1024;
            }
        };

        if ((status &ALTERA_AVALON_SPI_STATUS_RRDY_MSK) != 0)
        {
            //alt_u32 rxdata = IORD_ALTERA_AVALON_SPI_RXDATA(base);
            alt_u8 rxdata = base[ALTERA_AVALON_SPI_RXDATA_REG];

            //ȷֽ
            if (read_ignore > 0)
            {
                read_ignore--;
            }
            //ٽֽ
            else
            {
                *read_data++ = rxdata;
                //printf("%x ", rxdata);
            }
            credits++;

            //Ϻͽϣѭ
            if (read_ignore == 0 && read_data == read_end)
            {
                break;
            }
        }

    }
	delay=0x17f;
    // Wait until the interface has finished transmitting
    do
    {
        //status = IORD_ALTERA_AVALON_SPI_STATUS(base);
        status = base[ALTERA_AVALON_SPI_STATUS_REG];
		//wujc -s
		while(delay)
			delay--;
		//wujc -e
        //printf("tmt status=0x%x\n", status);
    }
    while ((status &ALTERA_AVALON_SPI_STATUS_TMT_MSK) == 0);

    // Clear SSO (release chipselect) unless the caller is going to keep using this chip
    if ((flags &ALT_AVALON_SPI_COMMAND_MERGE) == 0)
    {
        //IOWR_ALTERA_AVALON_SPI_CONTROL(base, 0);
        base[ALTERA_AVALON_SPI_CONTROL_REG] = 0;
        base[ALTERA_AVALON_SPI_CONTROL_REG + 1] = 0;
        base[ALTERA_AVALON_SPI_CONTROL_REG + 2] = 0;
        base[ALTERA_AVALON_SPI_CONTROL_REG + 3] = 0;
    }    
	
	// now we're done, so unlock the mutex
	if(pthread_mutex_unlock(&mutex) != 0) {
		fprintf(stderr, "unlock failed.\n");
		return -1;
	}

    return read_length;
}

//epcs_commands.c
static alt_u8 epcs_read_status_register(alt_u8 *base)
{
    //The Status Register may be read at any time, even while a Program, Erase or Write Status Register cycle is in progress. When one of these cycles is in progress, it is recommended to
    //check the Write In Progress (WIP) bit before sending a new instruction to the device. It is also possible to read the Status Register continuously,
    //You can read the status register at any time, even while a write or erase cycle is in progress. When one of these cycles is in progress, 
    //you can check the write in progress bit (bit 0 of the status register) before sending a new operation to the device.
    const alt_u8 rdsr = epcs_rdsr;
    alt_u8 status;
    alt_avalon_spi_command(base, 0, 1,  &rdsr, 1,  &status, 0);

    return status;
}

static int epcs_test_wip(alt_u8 *base)
{
    return epcs_read_status_register(base) &1;
    //WIP bitλ״̬Ĵһλ
    //WIP bit. The Write In Progress (WIP) bit indicates whether the memory is busy with a Write Status Register, Program or Erase cycle. 
    //When set to 1, such a cycle is in progress, when reset to 0 no such cycle is in progress.
}

static void epcs_await_wip_released(alt_u8 *base)
{
    //The Write In Progress (WIP) bit is provided in the Status Register so that the application program can monitor its value, polling it to establish when the previous Write cycle, Program cycle or Erase cycle is complete.
    /* Wait until the WIP bit goes low. */
    while (epcs_test_wip(base)){}
}

static void epcs_sector_erase(alt_u8 *base, alt_u32 offset)
{
    alt_u8 se[4];

    se[0] = epcs_se;
    se[1] = (offset >> 16) &0xFF;
    se[2] = (offset >> 8) &0xFF;
    se[3] = offset &0xFF;

    alt_avalon_spi_command(base, 0, sizeof(se) / sizeof(*se), se, 0, (alt_u8*)0, 0);

    epcs_await_wip_released(base);
    //While the Sector Erase cycle is in progress, the Status Register may be read to check the value of the Write In Progress (WIP) bit. 
    //The Write In Progress (WIP) bit is 1 during the self-timed Sector Erase cycle, and is 0 when it is completed. 
    //At some unspecified time before the cycle is completed, the Write Enable Latch (WEL) bit is reset.
}

static void epcs_bulk_erase(alt_u8 *base)
{
    const alt_u8 be = epcs_be;

    alt_avalon_spi_command(base, 0, 1,  &be, 0, (alt_u8*)0, 0);

    epcs_await_wip_released(base);
    //While the Bulk Erase cycle is in progress, the Status Register may be read to check the value of the Write In Progress (WIP) bit. 
    //The Write In Progress (WIP) bit is 1 during the self-timed Bulk Erase cycle, and is 0 when it is completed. 
    //At some unspecified time before the cycle is completed, the Write Enable Latch (WEL) bit is reset.
}

static alt_32 epcs_read_buffer(alt_u8 *base, int offset, alt_u8 *dest_addr, int length)
{
    alt_u8 read_command[4];

    read_command[0] = epcs_read;
    //3-byte address (A23-A0): λַǰλַں
    read_command[1] = (offset >> 16) &0xFF;
    read_command[2] = (offset >> 8) &0xFF;
    read_command[3] = offset &0xFF;

    /* I don't know why this is necessary, since I call await-wip after
     * all writing commands.
     */
    epcs_await_wip_released(base);

    alt_avalon_spi_command(base, 0, sizeof(read_command) / sizeof(*read_command), read_command, length, (alt_u8*)dest_addr, 0);

    return length;
}

static void epcs_write_enable(alt_u8 *base)
{
    //All instructions that modify data must be preceded by a Write Enable (WREN) instruction to set the Write Enable Latch (WEL) bit.
    //The Write Enable (WREN) instruction sets the Write Enable Latch (WEL) bit.
    //The Write Enable Latch (WEL) bit must be set prior to every Page Program (PP), Sector Erase (SE), Bulk Erase (BE) and Write Status Register (WRSR) instruction.
    const alt_u8 wren = epcs_wren;
    alt_avalon_spi_command(base, 0, 1,  &wren, 0, (alt_u8*)0, 0);

    //The Write Enable Latch (WEL) bit is reset under the following conditions:
    // Power-up
    // Write Disable (WRDI) instruction completion
    // Write Status Register (WRSR) instruction completion
    // Page Program (PP) instruction completion
    // Sector Erase (SE) instruction completion
    // Bulk Erase (BE) instruction completion

    //when set to 0 the internal Write Enable Latch is reset and no Write Status Register, Program or Erase instruction is accepted.
}

/* Write a partial or full page, assuming that page has been erased */
static alt_32 epcs_write_buffer(alt_u8 *base, int offset, const alt_u8 *src_addr, int length)
{
    alt_u8 pp[4+256];
    int i;

    //The Page Program (PP) instruction allows bytes to be programmed in the memory (changing bits from 1 to 0). 
    //Before it can be accepted, a Write Enable (WREN) instruction must previously have been executed.
    pp[0] = epcs_pp;
    pp[1] = (offset >> 16) &0xFF;
    pp[2] = (offset >> 8) &0xFF;
    pp[3] = offset &0xFF;

    //printf("length=%d\n",length);
    //ҪдflashֱӽӲPPĺ
    for (i=0; i<length; i++)
    {
        pp[i+4] = src_addr[i];
    }

    /* First, WREN */
    epcs_write_enable(base);

    /* Send the PP command */
    alt_avalon_spi_command(base, 0, sizeof(pp) / sizeof(*pp), pp, 0, (alt_u8*)0, 0/*ALT_AVALON_SPI_COMMAND_MERGE*/);
    //the Page Program (PP) instruction allows up to 256 bytes to be programmed at a time (changing bits from 1 to 0), provided that they lie in consecutive addresses on the same page of memory.

    /* Send the user's buffer */
    //alt_avalon_spi_command(base, 0, length, src_addr, 0, (alt_u8*)0, 0);
    

    /* Wait until the write is done.  This could be optimized -
     * if the user's going to go off and ignore the flash for
     * a while, its writes could occur in parallel with user code
     * execution.  Unfortunately, I have to guard all reads/writes
     * with wip-tests, to make that happen.
     */
    epcs_await_wip_released(base);
    //While the Page Program cycle is in progress, the Status Register may be read to check the value of the Write In Progress (WIP) bit. 
    //The Write In Progress (WIP) bit is 1 during the self-timed Page Program cycle, and is 0 when it is completed. 
    //At some unspecified time before the cycle is completed, the Write Enable Latch (WEL) bit is reset.

    return length;
}


static alt_u8 epcs_read_electronic_signature(alt_u8 *base)
{
    //The Deep Power-down mode is entered when the specific instruction (the Deep Powerdown (DP) instruction) is executed. 
    //The device remains in this mode until another specific instruction (the Release from Deep Power-down and Read Electronic Signature (RES) instruction) is executed.
    const alt_u8 res_cmd[] =
    {
        //The instruction code is followed by 3 dummy bytes
        epcs_res, 0, 0, 0
    };
    alt_u8 res;

    alt_avalon_spi_command(base, 0, sizeof(res_cmd) / sizeof(*res_cmd), res_cmd, 1,  &res, 0);

    return res;

}

//altera_avalon_epcs_flash_controller.c
static int alt_epcs_flash_query(t_alt_flash_epcs_dev *flash);

//  If you try to read beyond the end of the memory, you'll wrap around
// to the beginning.  Reads that start beyond the end of the memory are
// flagged as errors with EIO (is there a better error code?).

// This might be a candidate for optimization.  Precompute the last-address?
static int alt_epcs_test_address(alt_flash_dev *flash_info, alt_u32 offset)
{
    int ret_code = 0;
    // Error checking: if the block offset is outside of the memory, return -EIO.
    t_alt_flash_epcs_dev *f = (t_alt_flash_epcs_dev*)flash_info;

    const alt_u32 last_region_index = f->dev.number_of_regions - 1;
    alt_u32 last_device_address =  - 1+f->dev.region_info[last_region_index].offset + f->dev.region_info[last_region_index].region_size;

    if (offset > last_device_address)
    {
        // Someone tried to erase a block outside of this device's range.
        ret_code =  - EIO;
    }
    return ret_code;
}

/*
 *  If you try to read beyond the end of the memory, you'll wrap around
 *  to the beginning.  Reads that start beyond the end of the memory are
 *  flagged as errors with EIO (is there a better error code?).
 */
static int alt_epcs_flash_read(alt_flash_dev *flash_info, int offset, void *dest_addr, int length)
{
    int ret_code = 0;

    t_alt_flash_epcs_dev *f = (t_alt_flash_epcs_dev*)flash_info;

    ret_code = alt_epcs_test_address(flash_info, offset);

    if (ret_code >= 0)
    {
        ret_code = epcs_read_buffer(f->register_base, offset, dest_addr, length);

	    /* epcs_read_buffer returns the number of buffers read, but
	     * alt_epcs_flash_read returns 0 on success, <0 on failure.
	     */
        if (ret_code == length)
        {
            ret_code = 0;
        }
    }
    else
    {
        fprintf(stderr, "Reading EPCS flash error.\n");
    }
    
    return ret_code;
}

static int alt_epcs_flash_init(t_alt_flash_epcs_dev *flash)
{
    int ret_code = 0;

    // Set up function pointers and/or data structures as needed.
    ret_code = alt_epcs_flash_query(flash);

    //  Register this device as a valid flash device type
    if (ret_code < 0)
    {        
		fprintf(stderr, "Query epcs flash error.\n");
    }

    return ret_code;
}

static int alt_epcs_flash_query(t_alt_flash_epcs_dev *flash)
{
    int ret_code = 0;

	/* Decide if an epcs flash device is attached.
	 *  
	 * ret_code = -ENODEV = No device found!
	 */

	/* It may be necessary to unlock or lock sectors,
	 * reset the device, or whatever, to ensure that
	 * it's in a known working state.
	 */

	/* Send the RES command sequence */
	// get EPCS silicon ID
    flash->silicon_id = epcs_read_electronic_signature(flash->register_base);
    printf("Silicon ID: 0x%x\n", (unsigned int)(flash->silicon_id));

    // Fill in all device-specific parameters.
    if (flash->silicon_id == 0x16)
    // EPCS64
    {
        flash->dev.region_info[0].region_size = 64 * 1024 * 1024 / 8;
        flash->dev.region_info[0].number_of_blocks = 128;
        flash->dev.region_info[0].block_size = 65536;
    }
    else if (flash->silicon_id == 0x14)
    // EPCS16
    {
        flash->dev.region_info[0].region_size = 16 * 1024 * 1024 / 8;
        flash->dev.region_info[0].number_of_blocks = 32;
        flash->dev.region_info[0].block_size = 65536;
    }
    else if (flash->silicon_id == 0x13)
    // EPCS8
    {
        flash->dev.region_info[0].region_size = 8 * 1024 * 1024 / 8;
        flash->dev.region_info[0].number_of_blocks = 16;
        flash->dev.region_info[0].block_size = 65536;
    }
    else if (flash->silicon_id == 0x12)
    // EPCS4
    {
        flash->dev.region_info[0].region_size = 4 * 1024 * 1024 / 8; // = 65536 * 8 = 524288
        flash->dev.region_info[0].number_of_blocks = 8;
        flash->dev.region_info[0].block_size = 65536;
//        printf("Find the Serial Configuration Device: EPCS4\n");
    }
    else if (flash->silicon_id == 0x10)
    // EPCS1
    {
        flash->dev.region_info[0].region_size = 1 * 1024 * 1024 / 8;
        flash->dev.region_info[0].number_of_blocks = 4;
        flash->dev.region_info[0].block_size = 32768;
    }
    else
    {
        ret_code =  - ENODEV; // No known device found!
    }

    flash->size_in_bytes = flash->dev.region_info[0].region_size;
    flash->dev.number_of_regions = 1;
    flash->dev.region_info[0].offset = 0;
    flash->page_size = 256;

    // Consider clearing all BP bits here.
    return ret_code;
}

static int alt_epcs_flash_memcmp(alt_flash_dev *flash_info, const void *src_buffer, int offset, size_t n)
{
  /*
   * Compare chunks of memory at a time, for better serial-flash
   * read efficiency.
   */
    alt_u8 chunk_buffer[32];
    const alt_u32 chunk_size = sizeof(chunk_buffer) / sizeof(*chunk_buffer);
    int current_offset = 0;

    while (n > 0)
    {
        int this_chunk_size = n > chunk_size ? chunk_size : n;
        int this_chunk_cmp;

        if (alt_epcs_flash_read(flash_info, offset + current_offset, chunk_buffer, this_chunk_size) < 0)
        {
            // If the read fails, I'm not sure what the appropriate action is. Compare success seems wrong, so make it compare fail.
            return  - 1;
        }

        // Compare this chunk against the source memory buffer.
        this_chunk_cmp = memcmp(&((unsigned char*)(src_buffer))[current_offset], chunk_buffer, this_chunk_size);
        if (this_chunk_cmp)
        {
            return this_chunk_cmp;
        }

        n -= this_chunk_size;
        current_offset += this_chunk_size;
    }

    // If execution made it to this point, compare is successful.
    return 0;
}

// Program the data into the flash at the selected address.

// Restrictions - For now this function will program the sectors it needs,
// if it needs to erase a sector it will. If you wish to have all the contents
// of the sector preserved you the user need to be aware of this and read out
// the contents of that sector and add it to the data you wish to program.
// The reasoning here is that sectors can be very large eg. 64k which is a
// large buffer to tie up in our programming library, when not all users will
// want that functionality.
static int alt_epcs_flash_write(alt_flash_dev *flash_info, int offset, const void *src_addr, int length)
{
    int ret_code = 0;
    int i, j;
    int data_to_write;
    int current_offset;

    // First and foremost which sectors are affected?
    for (i = 0; i < flash_info->number_of_regions; i++)
    {
        // Is it in this erase block region?
        if ((offset >= flash_info->region_info[i].offset) && (offset < (flash_info->region_info[i].offset + flash_info->region_info[i].region_size)))
        {
            current_offset = flash_info->region_info[i].offset;

            for (j = 0; j < flash_info->region_info[i].number_of_blocks; j++)
            {
                if ((offset >= current_offset) && (offset < (current_offset + flash_info->region_info[i].block_size)))
                {
                    // Check if the contents of the block are different from the data we wish to put there
                    data_to_write = (current_offset + flash_info->region_info[i].block_size - offset);
                    data_to_write = MIN(data_to_write, length);

                    if (alt_epcs_flash_memcmp(flash_info, src_addr, offset, data_to_write))
                    {
                        ret_code = (*flash_info->erase_block)(flash_info, current_offset);

                        if (!ret_code)
                        {
                            ret_code = (*flash_info->write_block)(flash_info,
                            //current_offset,
                            offset, src_addr, data_to_write);
                        }
                    }

                    // Was this the last block?
                    if ((length == data_to_write) || ret_code)
                    {
                        goto finished;
                    }

                    length -= data_to_write;
                    offset = current_offset + flash_info->region_info[i].block_size;
                    src_addr = (alt_u8*)src_addr + data_to_write;
                }
                current_offset += flash_info->region_info[i].block_size;
            }
        }
    }

    finished: 
		return ret_code;
}

// Pass the table of erase blocks to the user
static int alt_epcs_flash_get_info(alt_flash_fd *fd, flash_region **info, int *number_of_regions)
{
    int ret_code = 0;

    alt_flash_dev *flash = (alt_flash_dev*)fd;

    *number_of_regions = flash->number_of_regions;

    if (!flash->number_of_regions)
    {
        ret_code =  - EIO;
    }
    else if (flash->number_of_regions > ALT_MAX_NUMBER_OF_FLASH_REGIONS)
    {
        ret_code =  - ENOMEM;
    }
    else
    {
        *info = &flash->region_info[0];
    }

    return ret_code;
}


/*
 *
 * Erase the selected erase block ("sector erase", from the POV
 * of the EPCS data sheet).
 */
static int alt_epcs_flash_erase_block(alt_flash_dev *flash_info, int block_offset)
{
    int ret_code = 0;
    t_alt_flash_epcs_dev *f = (t_alt_flash_epcs_dev*)flash_info;

    ret_code = alt_epcs_test_address(flash_info, block_offset);

    if (ret_code >= 0)
    {
        // Execute a WREN instruction
        epcs_write_enable(f->register_base);
        //The Erase instruction must be preceded by a Write Enable (WREN) instruction.

        // Send the Sector Erase command, whose 3 address bytes are anywhere within the chosen sector.
        epcs_sector_erase(f->register_base, block_offset);
    }
    else
    {
        fprintf(stderr, "Erasing epcs flash error.\n");
    }
    return ret_code;
}

static void alt_epcs_flash_erase_bulk(alt_flash_dev *flash_info)
{
    t_alt_flash_epcs_dev *f = (t_alt_flash_epcs_dev*)flash_info;

    // Execute a WREN instruction
    epcs_write_enable(f->register_base);
    //The Erase instruction must be preceded by a Write Enable (WREN) instruction.

    // Send the Bulk Erase command
    epcs_bulk_erase(f->register_base);
}

/* Write, assuming that someone has kindly erased the appropriate
 * sector(s).
 * Note: "block_offset" is the base of the current erase block.
 * "data_offset" is the absolute address (from the 0-base of this
 * device's memory) of the beginning of the write-destination.
 * This device has no need for "block_offset", but it's included for
 * function type compatibility.
 */
static int alt_epcs_flash_write_block(alt_flash_dev *flash_info,  //int block_offset,
	int data_offset, const void *data, int length)
{
	int ret_code;
	t_alt_flash_epcs_dev *f = (t_alt_flash_epcs_dev*)flash_info;

	int buffer_offset = 0;
	int length_of_current_write;
	ret_code = alt_epcs_test_address(flash_info, data_offset);

	if (ret_code >= 0)
	{
		// "Block" writes must be broken up into the page writes that the device understands.  Partial page writes are allowed.
		while (length)
		{
			int next_page_start = (data_offset + f->page_size) &~(f->page_size - 1);
			length_of_current_write = MIN(length, next_page_start - data_offset);
			//~(f->page_size - 1)=0xffff ff00
			//printf("next_page_start=%d\n", next_page_start);
			//printf("length_of_current_write=%d\n", length_of_current_write);

			//һд256ֽ
			epcs_write_buffer(f->register_base, data_offset, &((const alt_u8*)data)[buffer_offset], length_of_current_write);

			//鳤
			length -= length_of_current_write;
			//ʼλ
			buffer_offset += length_of_current_write;
			//ƶдλõһҳĿͷ
			data_offset = next_page_start;
		}
	}
	else
	{
		fprintf(stderr, "Writing EPCS flash error.\n");
	}

	return ret_code;
}

static int epcs_flash_destroy(void)
{
	int i;
	t_alt_flash_epcs_dev *flash_epcs_dev;
	
	if(popvx_pci_dev) {
		for(i = 0; i < max_epcs_dev; i ++) {
			flash_epcs_dev = popvx_pci_dev[i].epcs_flash_dev;
			if (flash_epcs_dev) {
				free(flash_epcs_dev);
				flash_epcs_dev = NULL;
			}
		}
		free(popvx_pci_dev);
		popvx_pci_dev = NULL;
	}

	menu_destroy_items(pmenu_epcs_flash);
	
	return 0;
}

static unsigned char epcs_modify_interrupt_pin(int pin_chg)
{
	int i;
	char change_pin_flag = 0;
	unsigned char pin[4];
	unsigned char ch = 0xFF;

	if (pin_chg == PIN_CHANGED_ENABLE) {
		while(1) {
	//		printf("Do you want to change the interrupt pin (y/[n])?");
			printf("Do you want to just change the interrupt pin (y/n)? ");
			setbuf(stdin, NULL);
			memset(pin, 0, 4);
			
			i = 0;
			while((ch = getchar()) != '\n') {
				pin[i++] = ch;
				if (i == 4) {
					break;
				}
			}

			if (pin[0] == '\0') {
	//			fprintf(stderr, "\nError: Invalid option.\n");
				continue;
			}
			if (pin[0] == 'y' || pin[0] == 'Y') {
				change_pin_flag = 1;
				break;
			} else if (pin[0] == 'n' || pin[0] == 'N') {
				return 0xFF;
			}
		}

		if (1 == change_pin_flag) {
			while(1) {
				printf("Please input the pin number ([A|B|C|D]): ");
				setbuf(stdin, NULL);
				memset(pin, 0, 4);
				
				i = 0;
				while((ch = getchar()) != '\n') {
					pin[i++] = ch;
					if (i == 4) {
						break;
					}
				}

				if (pin[0] == '\0') {
	//				fprintf(stderr, "\nError: Invalid option.\n");
					continue;
				}
				if (pin[0] == 'A' || pin[0] == 'B' || pin[0] == 'C' || pin[0] == 'D' ||
					pin[0] == 'a' || pin[0] == 'b' || pin[0] == 'c' || pin[0] == 'd') {
					change_pin_flag = 2;
					break;
				}
			}
		}

		if (2 == change_pin_flag) {
			switch(pin[0]) {
				case 'A':
				case 'a':
					ch = 0x01; // or 0xFF
					break;
				case 'B':
				case 'b':
					ch = 0x02;
					break;
				case 'C':
				case 'c':
					ch = 0x03;
					break;
				case 'D':
				case 'd':
					ch = 0x04;
					break;
				default:
					ch = 0xFF;
			}
		}
	}

	return ch;
}


/************************************************************************
 * vendor_iddevice_idPCI豸豸
 * ̬epcs_flash_dev豸
 * ̬˵
 ************************************************************************/
static int epcs_flash_scan(int menu_oper_item)
{
	struct pci_access *pacc = NULL;
	struct pci_dev *dev;
	unsigned char pci_idx = 0;
	unsigned char idx = 0;
	t_alt_flash_epcs_dev * tmp = NULL;
	unsigned char *baseoffset;		
	int i;

	int flag = 0;

	//--------------------------------------------------------------------------
	/* Get the pci_access structure */
	pacc = pci_alloc(); 

	/* Set all options you want -- here we stick with the defaults */
	/* Initialize the PCI library */
	pci_init(pacc); 

	/* We want to get the list of devices */
	pci_scan_bus(pacc); 

	//--------------------------------------------------------------------------
	// vendor_iddevice_idPCI豸 max_epcs_dev
	// һɨopenvovxİ忨ð忨Ŀ
	/* Iterate over all devices */
	for (dev = pacc->devices; dev; dev = dev->next)
	{
		/* Fill in header info we need */
		pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_BASES | PCI_FILL_CLASS);

		if (pci_idx > EPCS_MAX_DEVICE)
		{
			fprintf(stderr, "Max epcs device error.\n");
			/* Close everything */
			pci_cleanup(pacc); 
			exit(1);
		}		// check the vendor id and device id

		if (((G400_VENDOR_ID  == dev->vendor_id) && (G400_DEVICE_ID  == dev->device_id)) ||
			((A2410_VENDOR_ID == dev->vendor_id) && (A2410_DEVICE_ID == dev->device_id)) ||
			((A1610_VENDOR_ID == dev->vendor_id) && (A1610_DEVICE_ID == dev->device_id)) ||
			((A810_VENDOR_ID  == dev->vendor_id) && (A810_DEVICE_ID  == dev->device_id)) ||
			((D830_VENDOR_ID  == dev->vendor_id) && (D830_DEVICE_ID  == dev->device_id)) ||
			((D430_VENDOR_ID  == dev->vendor_id) && (D430_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D430_VENDOR_ID  == dev->vendor_id) && (OPVX_D430_DEVICE_ID  == dev->device_id)) ||
			((D230_VENDOR_ID  == dev->vendor_id) && (D230_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D230_VENDOR_ID  == dev->vendor_id) && (OPVX_D230_DEVICE_ID  == dev->device_id)) ||
			((D130_VENDOR_ID  == dev->vendor_id) && (D130_DEVICE_ID  == dev->device_id)) ||			
			((D410G2_VENDOR_ID  == dev->vendor_id) && (D410G2_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D410G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D410G2_DEVICE_ID  == dev->device_id)) ||
			((D210G2_VENDOR_ID  == dev->vendor_id) && (D210G2_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D210G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D210G2_DEVICE_ID  == dev->device_id)) ||
			((X200_VENDOR_ID  == dev->vendor_id) && (X200_DEVICE_ID  == dev->device_id)) ||
			((D115G2_VENDOR_ID  == dev->vendor_id) && (D115G2_DEVICE_ID  == dev->device_id)))
		{
			pci_idx ++;
		}

	}
		
	max_epcs_dev = pci_idx;
	if(0 == max_epcs_dev) {
		fprintf(stderr, "No OpenVox cards found.\n");
		/* Close everything */
		pci_cleanup(pacc); 
		pacc = NULL;
		exit(1);	
	}

	//--------------------------------------------------------------------------
	// ɨİ忨̬epcs_flash豸
	popvx_pci_dev = (t_opvx_pci_dev_s *)calloc(max_epcs_dev, sizeof(t_opvx_pci_dev_s));
	if (!popvx_pci_dev) {
		fprintf(stderr, "Unable to allocate memory.\n");
		pci_cleanup(pacc); 
		exit(-1);
	}
	

	//--------------------------------------------------------------------------
	// ڶɨopenvovxİ忨ð忨ĻϢ
	idx = 0;
	for (dev = pacc->devices; dev; dev = dev->next)
	{
		// Fill in header info we need
		pci_fill_info(dev, PCI_FILL_IDENT | PCI_FILL_BASES | PCI_FILL_CLASS);

		if (idx > EPCS_MAX_DEVICE)
		{
			epcs_flash_destroy();
			fprintf(stderr, "Max epcs device error.\n");
			pci_cleanup(pacc); 
			exit(1);
		}

		//----------------------------------------------------------------------
		// ð忨ĻϢ
		if (((G400_VENDOR_ID  == dev->vendor_id) && (G400_DEVICE_ID  == dev->device_id)) ||
			((A2410_VENDOR_ID == dev->vendor_id) && (A2410_DEVICE_ID == dev->device_id)) ||
			((A1610_VENDOR_ID == dev->vendor_id) && (A1610_DEVICE_ID == dev->device_id)) ||
			((A810_VENDOR_ID  == dev->vendor_id)  && (A810_DEVICE_ID == dev->device_id)) ||
			((D830_VENDOR_ID  == dev->vendor_id) && (D830_DEVICE_ID  == dev->device_id)) ||
			((D430_VENDOR_ID  == dev->vendor_id) && (D430_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D430_VENDOR_ID  == dev->vendor_id) && (OPVX_D430_DEVICE_ID  == dev->device_id)) ||
			((D230_VENDOR_ID  == dev->vendor_id) && (D230_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D230_VENDOR_ID  == dev->vendor_id) && (OPVX_D230_DEVICE_ID  == dev->device_id)) ||
			((D130_VENDOR_ID  == dev->vendor_id) && (D130_DEVICE_ID  == dev->device_id)) ||				
			((D410G2_VENDOR_ID  == dev->vendor_id) && (D410G2_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D410G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D410G2_DEVICE_ID  == dev->device_id)) ||
			((D210G2_VENDOR_ID  == dev->vendor_id) && (D210G2_DEVICE_ID  == dev->device_id)) ||
			((OPVX_D210G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D210G2_DEVICE_ID  == dev->device_id)) ||
			((X200_VENDOR_ID  == dev->vendor_id) && (X200_DEVICE_ID  == dev->device_id)) ||
			((D115G2_VENDOR_ID  == dev->vendor_id) && (D115G2_DEVICE_ID  == dev->device_id)))
		{
			// set epcs flash dev
			popvx_pci_dev[idx].mem_base 		= dev->base_addr[0];
			popvx_pci_dev[idx].vendor_id 		= dev->vendor_id;
			popvx_pci_dev[idx].device_id 		= dev->device_id;
			popvx_pci_dev[idx].id 				= idx;

			popvx_pci_dev[idx].domain 			= dev->domain;
			popvx_pci_dev[idx].bus 				= dev->bus;
			popvx_pci_dev[idx].dev 				= dev->dev;
			popvx_pci_dev[idx].func 			= dev->func;
			popvx_pci_dev[idx].irq 				= dev->irq;
			

			// get interrupt pin
			popvx_pci_dev[idx].pin 				= pci_read_byte(dev, 0x3d);
		}


		//----------------------------------------------------------------------
		// ð忨ĲϢ
		// check the vendor id and device id
		if ((G400_VENDOR_ID  == dev->vendor_id) && (G400_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "G400P";
			popvx_pci_dev[idx].card_type 		            = OPVX_G400P;
			popvx_pci_dev[idx].baseoff 			            = G400_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = G400_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= G400_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= G400_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_DISABLE;
			flag = 1;
		} else if ((A2410_VENDOR_ID == dev->vendor_id) && (A2410_DEVICE_ID == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "A2410";			
			popvx_pci_dev[idx].card_type 		            = OPVX_A2410P;
			popvx_pci_dev[idx].baseoff 			            = A2410_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = A2410_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= A2410_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= A2410_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((A1610_VENDOR_ID == dev->vendor_id) && (A1610_DEVICE_ID == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "A1610";			
			popvx_pci_dev[idx].card_type 		            = OPVX_A1610;
			popvx_pci_dev[idx].baseoff 			            = A1610_EPCS_BASE;
			popvx_pci_dev[idx].register_offset	            = A1610_EPCS_OFFSET;
			popvx_pci_dev[idx].c3reload_baseoff 			= A1610_C3RELOAD_BASE;
			popvx_pci_dev[idx].c3reload_register_offset		= A1610_C3RELOAD_OFFSET;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((A810_VENDOR_ID == dev->vendor_id) && (A810_DEVICE_ID == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "A810";			
			popvx_pci_dev[idx].card_type 		            = OPVX_A810;
			popvx_pci_dev[idx].baseoff 			            = A810_EPCS_BASE;
			popvx_pci_dev[idx].register_offset	            = A810_EPCS_OFFSET;
			popvx_pci_dev[idx].c3reload_baseoff 			= A810_C3RELOAD_BASE;
			popvx_pci_dev[idx].c3reload_register_offset		= A810_C3RELOAD_OFFSET;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D830_VENDOR_ID  == dev->vendor_id) && (D830_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "Digium";
			popvx_pci_dev[idx].card_name 		            = "D830";			
			popvx_pci_dev[idx].card_type 		            = WC_D830P;
			popvx_pci_dev[idx].baseoff 			            = D830_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D830_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D830_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D830_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D430_VENDOR_ID  == dev->vendor_id) && (D430_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "Digium";
			popvx_pci_dev[idx].card_name 		            = "D430";			
			popvx_pci_dev[idx].card_type 		            = WC_D430P;
			popvx_pci_dev[idx].baseoff 			            = D430_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D430_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D430_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D430_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((OPVX_D430_VENDOR_ID  == dev->vendor_id) && (OPVX_D430_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D430";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D430P;
			popvx_pci_dev[idx].baseoff 			            = OPVX_D430_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = OPVX_D430_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= OPVX_D430_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= OPVX_D430_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D230_VENDOR_ID  == dev->vendor_id) && (D230_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "Digium";
			popvx_pci_dev[idx].card_name 		            = "D230";			
			popvx_pci_dev[idx].card_type 		            = WC_D230P;
			popvx_pci_dev[idx].baseoff 			            = D230_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D230_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D230_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D230_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((OPVX_D230_VENDOR_ID  == dev->vendor_id) && (OPVX_D230_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D230";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D230P;
			popvx_pci_dev[idx].baseoff 			            = OPVX_D230_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = OPVX_D230_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= OPVX_D230_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= OPVX_D230_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D130_VENDOR_ID  == dev->vendor_id) && (D130_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D130";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D130P;
			popvx_pci_dev[idx].baseoff 			            = D130_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D130_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D130_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D130_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D410G2_VENDOR_ID  == dev->vendor_id) && (D410G2_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "Digium";
			popvx_pci_dev[idx].card_name 		            = "D410G2";			
			popvx_pci_dev[idx].card_type 		            = WC_D410G2;                     
			popvx_pci_dev[idx].baseoff 			            = D410G2_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D410G2_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D410G2_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D410G2_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((OPVX_D410G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D410G2_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D410G2";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D410G2;                     
			popvx_pci_dev[idx].baseoff 			            = OPVX_D410G2_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = OPVX_D410G2_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= OPVX_D410G2_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= OPVX_D410G2_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D210G2_VENDOR_ID  == dev->vendor_id) && (D210G2_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "Digium";
			popvx_pci_dev[idx].card_name 		            = "D210G2";			
			popvx_pci_dev[idx].card_type 		            = WC_D210G2;
			popvx_pci_dev[idx].baseoff 			            = D210G2_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D210G2_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D210G2_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D210G2_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((OPVX_D210G2_VENDOR_ID  == dev->vendor_id) && (OPVX_D210G2_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D210G2";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D210G2;
			popvx_pci_dev[idx].baseoff 			            = OPVX_D210G2_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = OPVX_D210G2_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= OPVX_D210G2_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= OPVX_D210G2_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((D115G2_VENDOR_ID  == dev->vendor_id) && (D115G2_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "D115G2";			
			popvx_pci_dev[idx].card_type 		            = OPVX_D115G2;
			popvx_pci_dev[idx].baseoff 			            = D115G2_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = D115G2_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= D115G2_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= D115G2_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else if ((X200_VENDOR_ID  == dev->vendor_id) && (X200_DEVICE_ID  == dev->device_id)) {
			popvx_pci_dev[idx].manufacturer 	            = "OpenVox";
			popvx_pci_dev[idx].card_name 		            = "X200";			
			popvx_pci_dev[idx].card_type 		            = OPVX_X200;
			popvx_pci_dev[idx].baseoff 			            = X200_EPCS_BASE ;
			popvx_pci_dev[idx].register_offset	            = X200_EPCS_OFFSET ;
			popvx_pci_dev[idx].c3reload_baseoff 			= X200_C3RELOAD_BASE ;
			popvx_pci_dev[idx].c3reload_register_offset		= X200_C3RELOAD_OFFSET ;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_ENABLE;
			flag = 1;
		} else {
			popvx_pci_dev[idx].manufacturer 	            = NULL;
			popvx_pci_dev[idx].card_name 		            = NULL;			
			popvx_pci_dev[idx].card_type 		            = OPVX_NONE;
			popvx_pci_dev[idx].baseoff 			            = 0;
			popvx_pci_dev[idx].register_offset	            = 0;
			// c3reload
			popvx_pci_dev[idx].c3reload_baseoff 			= 0;
			popvx_pci_dev[idx].c3reload_register_offset		= 0;
			popvx_pci_dev[idx].pin_chg			            = PIN_CHANGED_DISABLE;
			flag = 0;
		}
		
		//----------------------------------------------------------------------
		// ÿĶӦϵ
		card_indexes[idx] = popvx_pci_dev[idx].card_type;
		if (OPVX_NONE != card_indexes[idx]) {
			card_numbers[card_indexes[idx]] ++;
			card_menu_indexes[card_indexes[idx]] = idx;
		}

		//----------------------------------------------------------------------
		// ð忨flashϢṹ
		if (flag == 1)
		{
			tmp = (t_alt_flash_epcs_dev *)(malloc(sizeof(t_alt_flash_epcs_dev)));
			memset(tmp, 0, sizeof(t_alt_flash_epcs_dev));

			if(tmp) {
				tmp->dev.llist.next		= NULL;
				tmp->dev.llist.previous	= NULL;
				tmp->dev.name			= EPCS_CONTROLLER0_NAME;
				tmp->dev.open			= NULL;
				tmp->dev.close			= NULL;
				tmp->dev.write			= alt_epcs_flash_write;
				tmp->dev.read			= alt_epcs_flash_read;
				tmp->dev.get_info		= alt_epcs_flash_get_info;
				tmp->dev.erase_block	= alt_epcs_flash_erase_block;
				tmp->dev.write_block	= alt_epcs_flash_write_block;
				baseoffset = (unsigned char)&popvx_pci_dev[idx].baseoff;

				if(OPVX_NONE == popvx_pci_dev[idx].card_type) {
					fprintf(stderr, "Unknown card type\n");
				} else {
					//-tmp->dev.base_addr	= (void *)(popvx_pci_dev[idx].baseoff);
					//-tmp->dev.base_addr	= (unsigned char)(*baseoffset);
					tmp->dev.base_addr	= baseoffset;
					//-tmp->register_base 	= (unsigned char *)(popvx_pci_dev[idx].baseoff + popvx_pci_dev[idx].register_offset);	
					tmp->register_base 	= (baseoffset + popvx_pci_dev[idx].register_offset);	
				}
	
				popvx_pci_dev[idx].epcs_flash_dev 	= tmp;
				idx ++;
			} else {
				fprintf(stderr, "Unable to allocate memory.\n");			
				epcs_flash_destroy();
				pci_cleanup(pacc); 
				exit(1);
			}
		}
		
	}

#ifdef DEBUG
	int k;
	for (k = 0; k < idx; k ++) {
		printf("card_indexs[%d] = %d\n", k, card_indexes[k]);
	}
	for (k = 0; k < OPVX_END; k ++) {
		printf("card_numbers[%d] = %d\n", k, card_numbers[k]);
	}
	for (k = 0; k < OPVX_END; k ++) {
		printf("card_menu_indexes[%d] = %d\n", k, card_menu_indexes[k]);
	}
#endif

	//--------------------------------------------------------------------------
	// ֤ɨǷһ
	if (idx != pci_idx) {
		fprintf(stderr, "Recheck pci devices error.\n");					
		fprintf(stderr, "Recheck pci devices error. idx=%x pci_idx=%x\n", idx, pci_idx);					
		epcs_flash_destroy();
		exit(1);
	}

	//--------------------------------------------------------------------------
	// ɨ赽Ŀ࣬ڶ̬ɲ˵
	for (i = 1; i < OPVX_END; i ++) {
		if (card_numbers[i]) {
			card_type_number ++;
		}
	}
	
	// -------------------------------------------------------------------------
	// ̬˵һ˵ΪһθеĿ
	pmenu_epcs_flash = (t_menu_item *)calloc((MENU_CARD_UPDATE == menu_oper_item) ? \
	                   (max_epcs_dev + card_type_number): max_epcs_dev, \
	                   sizeof(t_menu_item));
	if (!pmenu_epcs_flash) {
		fprintf(stderr, "Unable to allocate memory!");
		epcs_flash_destroy();
		exit(-1);
	}

	// -------------------------------------------------------------------------
	// ò˵
	for (i = 0; i < pci_idx; i ++) {
		// Get Interrupt Pin
		u8 pin = popvx_pci_dev[i].pin;
		snprintf(pmenu_epcs_flash[i].name, MENU_MAXI_ITEM_LENGTH - 1, 
				"%d. %s %s (%04x:%02x:%02x.%d %04x:%04x pin %s IRQ %d)", 
				i + 1, 
				popvx_pci_dev[i].manufacturer,
				popvx_pci_dev[i].card_name,
				popvx_pci_dev[i].domain, popvx_pci_dev[i].bus, popvx_pci_dev[i].dev, popvx_pci_dev[i].func, 
				popvx_pci_dev[i].vendor_id, popvx_pci_dev[i].device_id,
				(pin == 1) ? "A" : (pin == 2) ? "B" : (pin == 3) ? "C": (pin == 4) ? "D" : "Undefined",
				popvx_pci_dev[i].irq
		);
		pmenu_epcs_flash[i].hotkey     = i + 1;
		pmenu_epcs_flash[i].device_idx = i;
	}
	
	
	// -------------------------------------------------------------------------
	// ò˵: еİ忨
	enum enum_card_type type;
	if (MENU_CARD_UPDATE == menu_oper_item) {
		for (i = 0, type = OPVX_NONE + 1; type < OPVX_END; type ++) {
			if (card_numbers[type]) {
//				printf("max_epcs_dev = %d, type = %d pci_idx = %d card_numbers[type] = %d\n",
//					    max_epcs_dev, type, pci_idx, card_numbers[type]);
				
				pmenu_epcs_flash[max_epcs_dev + i].hotkey = pci_idx + 1 + i;
				pmenu_epcs_flash[max_epcs_dev + i].device_idx            = card_menu_indexes[type];
#ifdef DEBUG
				printf("pmenu_epcs_flash[%d].device_idx = %d hotkey = %d\n",\
					     max_epcs_dev + i, pmenu_epcs_flash[max_epcs_dev + i].device_idx,
					     pmenu_epcs_flash[max_epcs_dev + i].hotkey);
#endif

				snprintf(pmenu_epcs_flash[max_epcs_dev + i].name, MENU_MAXI_ITEM_LENGTH - 1, 
						"%d. Update all %s %s", 
						pmenu_epcs_flash[max_epcs_dev + i].hotkey, 
						popvx_pci_dev[card_menu_indexes[type]].manufacturer,
						popvx_pci_dev[card_menu_indexes[type]].card_name
				);

				i ++;
			}
		}
	}


	/* Close everything */
	if (pacc)
	{
		pci_cleanup(pacc); 
	}

	/* OpenVox GSM PCI豸 */
	return pci_idx;
}


/************************************************************************
 * map_epcs4_reg
 ************************************************************************/
static int epcs_map_reg(t_opvx_pci_dev_s *cur_pci_dev)
{
    unsigned int base0, basepage, baseoff, *baseoffset;
    int fd;

 	if (!cur_pci_dev) {
		return -1;
	}

    if ((fd = open("/dev/mem", O_RDWR)) == 0)
    {
        fprintf(stderr, "open /dev/mem error.\n");
        return  - 1;
    }

    //ַ+PCIĴƫַ
    base0 = cur_pci_dev->mem_base + cur_pci_dev ->baseoff+ cur_pci_dev->register_offset;
    baseoff = base0 % getpagesize();
    basepage = base0 - baseoff;
    baseoffset = &baseoff;

//    printf("getpagesize() = %ld mem_base = 0x%.8X base0 = 0x%.8X basepage=0x%.8X  baseoff=0x%.8X\n", 
//			getpagesize(), opvx_pci_dev->mem_base, base0, basepage, baseoff);
// 		      getpagesize() = 4096 basepage=0xFEBF2000, baseoff=0x00000C00, destaddr=0xFEBF2C00
	cur_pci_dev->epcs_reg = NULL;
    if ((cur_pci_dev->epcs_reg = (unsigned char*)mmap(0,  /*ָļӦӳ䵽̿ռʼַһ㱻ָһָ룬ʱѡʼַں*/
    getpagesize(),  		/* ӳ䵽ý̵ַռֽӱӳļͷoffsetֽڿʼ */
    PROT_READ | PROT_WRITE, /* ڴķȨ */
    MAP_SHARED, fd,  		/* ӳ䵽̿ռļ֣һopen() */
    basepage 				/* һΪ0ʾļͷʼӳ */
    )) == MAP_FAILED)
     /* ķֵΪļӳ䵽̿ռĵֱַ̿ӲʼַΪֵЧַ */
    {
		cur_pci_dev->epcs_reg = NULL;
        fprintf(stderr, "Unable to mmap.\n");
        return  -1;
    }
	cur_pci_dev->base = (cur_pci_dev->epcs_reg + (*baseoffset));
	if (cur_pci_dev->epcs_flash_dev == NULL) {
		fprintf(stderr, "No epcs flash device.\n");
		close(fd);
		return -2;
	} else {
		cur_pci_dev->epcs_flash_dev->register_base = cur_pci_dev->base;
//printf("-------------wujc-debug: baseoff=0x%x base=0x%p register_base=0x%p\n", baseoff, cur_pci_dev->base, cur_pci_dev->epcs_flash_dev->register_base);
	}
	
    close(fd);
    
    return 0;
}

static int epcs_unmap_reg(t_opvx_pci_dev_s *cur_pci_dev)
{
	if (cur_pci_dev && cur_pci_dev->epcs_reg) {
    	munmap(cur_pci_dev->epcs_reg, getpagesize());
		cur_pci_dev->epcs_reg = NULL;
		return 0;
	}
	return -1;
}

/************************************************************************
 * c3reload
 ************************************************************************/
static int C3Reload_map_reg(t_opvx_pci_dev_s *cur_pci_dev)
{
    unsigned int base0, basepage, baseoff, *baseoffset;
    int fd;

 	if (!cur_pci_dev) {
		return -1;
	}
	
    if ((fd = open("/dev/mem", O_RDWR)) == 0)
    {
        fprintf(stderr, "open /dev/mem error.\n");
        return  - 1;
    }

    //ַ+PCIĴƫַ
    base0 = cur_pci_dev->mem_base + cur_pci_dev ->c3reload_baseoff+ cur_pci_dev->c3reload_register_offset;
    baseoff = base0 % getpagesize();
    basepage = base0 - baseoff;

    baseoffset = &baseoff;

//    printf("getpagesize() = %ld basepage=0x%.8X, baseoff=0x%.8X, destaddr=0x%.8X\n", getpagesize(), basepage, opvx_pci_dev->baseoff, basepage + opvx_pci_dev->baseoff);
// 		      getpagesize() = 4096 basepage=0xFEBF2000, baseoff=0x00000C00, destaddr=0xFEBF2C00

	// mmap
	cur_pci_dev->c3reload_reg = NULL;
    if ((cur_pci_dev->c3reload_reg = (unsigned char*)mmap(0,  /*ָļӦӳ䵽̿ռʼַһ㱻ָһָ룬ʱѡʼַں*/
    getpagesize(),  		/* ӳ䵽ý̵ַռֽӱӳļͷoffsetֽڿʼ */
    PROT_READ | PROT_WRITE, /* ڴķȨ */
    MAP_SHARED, fd,  		/* ӳ䵽̿ռļ֣һopen() */
    basepage 				/* һΪ0ʾļͷʼӳ */
    )) == MAP_FAILED)
     /* ķֵΪļӳ䵽̿ռĵֱַ̿ӲʼַΪֵЧַ */
    {
        fprintf(stderr, "Unable to mmap.\n");
        return  -1;
    }

	//-cur_pci_dev->c3reload_base = (unsigned int)(cur_pci_dev->c3reload_reg + baseoff);
	cur_pci_dev->c3reload_base = (cur_pci_dev->c3reload_reg + (*baseoffset));
	
    close(fd);
    
    return 0;
}

static int C3Reload_unmap_reg(t_opvx_pci_dev_s *cur_pci_dev)
{
 	if (!cur_pci_dev) {
		return -1;
	}
	if (cur_pci_dev->c3reload_reg)
	{
	    munmap(cur_pci_dev->c3reload_reg, getpagesize());
		return 0;
	}

	return -1;
}

static int C3Reload_run(t_opvx_pci_dev_s *cur_pci_dev)
{
	unsigned int *base;

 	if (!cur_pci_dev) {
		return -1;
	}
	
	C3Reload_map_reg(cur_pci_dev);
	
	if(pthread_mutex_lock(&mutex) != 0) {
		fprintf(stderr, "Lock failed.\n");
		return -1;
	}

	base = (unsigned int *)cur_pci_dev->c3reload_base;
    //-base[0] = 0x01010000;//&0xff;
	if(cur_pci_dev->card_name == "X200")
		*base = 0x01810080;//&0xff;
	else
		*base = 0x01010000;//&0xff;
	
	// now we're done, so unlock the mutex
	if(pthread_mutex_unlock(&mutex) != 0) {
		fprintf(stderr, "unlock failed.\n");
		return -1;
	}
	
	C3Reload_unmap_reg(cur_pci_dev);
		
	return 0;
}

//NIOS2ǸλǰX86ǵλǰҪNIOS2ֽ任֮Ȼ
static unsigned char exchange(unsigned char abcd)
{
    unsigned char dcba = 0;
    unsigned char high, low;
    unsigned char high_mask = 0x80, low_mask = 0x01;
    int i;

    //flashд0xff0x00ôֱӷ
    if ((abcd == 0xff) || (abcd == 0x00))
        return abcd;
    
    for (i = 0; i < 4; i++)
    {
        //Ƚ7λ͵0λ
        high = abcd &high_mask; //1000 0000
        low = abcd &low_mask; //0000 0001
        high >>= 7-i * 2;
        low <<= 7-i * 2;
        dcba |= high | low;
        //printf("jic=%X\n", jic);

        high_mask >>= 1;
        low_mask <<= 1;
    }
    return dcba;
}

/*  
    getFileSize(char * strFileName)   
    ܣȡָļĴС  
      
        strFileName (char *)ļ  
    ֵ  
        size (int)ļС  
 */  
static int getFileSize(char * strFileName)
{   
    FILE * fp = fopen(strFileName, "r");   
    fseek(fp, 0L, SEEK_END);   
    int size = ftell(fp);   
    fclose(fp);   
    return size;   
}

static int epcs_flash_init(void)
{
	int i;
	
	// ٶ̬ڴ
	if (max_epcs_dev)
	{
		epcs_flash_destroy();
		menu_destroy_items(pmenu_epcs_flash);
	}

	// ʼȫֱ
	max_epcs_dev = 0;
	for (i = 0; i < OPVX_END; i ++) {
		card_menu_indexes[i] = -1;
	}
	return 0;
}
#if 0
/************************************************************************
 * źŴ
 ************************************************************************/
static void ctrl_c_op(int signo)
{
	return;
}
#endif
static int epcs_flash_backup_rpd(t_opvx_pci_dev_s *cur_pci_dev)
{
	char filename[32];
	int ret_code;
	FILE *fp;
	int i;
	unsigned char dest[FLASH_SIZE];
	unsigned char rpd;

	if (!cur_pci_dev || !cur_pci_dev->epcs_flash_dev) {
		fprintf(stderr, "No EPCS controller found.\n");	
		return -1;
	}

	struct alt_flash_epcs_dev *epcs = cur_pci_dev->epcs_flash_dev;
	struct alt_flash_dev *dev		= &(epcs->dev);
	
	//ݾɵflash
	printf("Please input the filename for backup: ");
	scanf("%s", filename);

	// ½ļ
	if ((fp = fopen(filename, "wb")) == NULL)
	{
		printf("Cannot open file: %s\n", filename);
		return  -1;
	}

	struct sigaction act;
	//-act.sa_handler=ctrl_c_op;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}

	//ýȰٷֱ
	printf("Reading epcs flash...\n");
	syslog(LOG_INFO, "Reading epcs flash...\n");

	// flash
	ret_code = alt_epcs_flash_read(dev, 0, dest, FLASH_SIZE);
	if (ret_code)
	{
		fprintf(stderr, "Reading epcs flash error.\n");
		return -1;
	}

	// ֽڷתд뱸ļ
	for (i = 0; i < FLASH_SIZE; i++)
	{
			rpd = exchange(dest[i]);
			fputc(rpd, fp);

	}

	fclose(fp);

	act.sa_handler=NULL;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}

	printf("Reading epcs flash successfully\n");
	syslog(LOG_INFO, "Reading epcs flash successfully\n");

	return 0;
}


/*******************************************************************************
 *
 * ö
 *
 * return value:
      -1 : sigaction error
      -2 : epcs flash operation error
       0 : öγɹ
       1 : ö
 ******************************************************************************/
static int epcs_flash_update_config(t_opvx_pci_dev_s *cur_pci_dev)
{
	int ret_code;
	unsigned char ch;
	unsigned char dest[EPCS_CONFIG_PAGE_SIZE];

	if (!cur_pci_dev || !cur_pci_dev->epcs_flash_dev) {
		fprintf(stderr, "No EPCS controller found.\n");	
		return -1;
	}

	struct alt_flash_epcs_dev *epcs = cur_pci_dev->epcs_flash_dev;
	struct alt_flash_dev *dev		= &(epcs->dev);
		
	// get interrupt pin code
	// FF: ʾıinterrupt pin
	ch = epcs_modify_interrupt_pin(cur_pci_dev->pin_chg);
	if (0xFF == ch) {
		return 1;
	}

	struct sigaction act;
	//-act.sa_handler=ctrl_c_op;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}

	// ȡöϢ
	printf("Reading epcs flash...\n");
	syslog(LOG_INFO, "Reading epcs flash...\n");
	ret_code = alt_epcs_flash_read(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
	if (ret_code)
	{
		fprintf(stderr, "Reading epcs flash error.\n");
		return -2;
	}
	printf("Reading epcs flash done\n");
	syslog(LOG_INFO, "Reading epcs flash done\n");


	// ö
	printf("Erasing epcs flash...\n");
	syslog(LOG_INFO, "Erasing epcs flash...\n");
	alt_epcs_flash_erase_block(dev, EPCS_CONFIG_PAGE_OFFSET);
	printf("Erasing epcs flash done\n");
	syslog(LOG_INFO, "Erasing epcs flash done\n");

	// дµö
	printf("Writing epcs flash...\n");
	syslog(LOG_INFO, "Writing epcs flash...\n");

	// change interrupt pin register value
	dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_INTS] = ch ;
	
	ret_code = alt_epcs_flash_write_block(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
	if (ret_code)
	{
		fprintf(stderr, "Writng epcs flash error.\n");
		return -2;
	}

	act.sa_handler=NULL;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "Error\n");
		return -1;
	}

	printf("Writing epcs flash successfully\n");
	syslog(LOG_INFO, "Writing epcs flash successfully\n");

	return 0;
}


static int epcs_flash_update_rpd(t_opvx_pci_dev_s *cur_pci_dev)
{
	char filename[32];
	unsigned char rpd;
	int ret_code;
	unsigned char source[FLASH_SIZE];
	unsigned char ch;
	FILE *fp;
	int i;
	int res;
	struct stat buf;
	mode_t mode;
	unsigned short vendorid,devid;
	unsigned char char_input[4];
	
	if (!cur_pci_dev || !cur_pci_dev->epcs_flash_dev) {
		fprintf(stderr, "No EPCS controller found.\n");	
		return -1;
	}

	struct alt_flash_epcs_dev *epcs = cur_pci_dev->epcs_flash_dev;
	struct alt_flash_dev *dev		= &(epcs->dev);

#ifdef DEBUG
	printf(DEBUGFMT "Card Name = %s Card Type = %d\n", DEBUGARGS, cur_pci_dev->card_name, cur_pci_dev->card_type);
#endif
		
	res = epcs_flash_update_config(cur_pci_dev);
	if (1 != res) {
		return 0;
	}
		
	//printf("\t---------------Debug: VendorID:0x%x  DeviceID:0x%x FW_SIZE:0x%x \n", cur_pci_dev->vendor_id, cur_pci_dev->device_id, FLASH_SIZE);
	printf("Please input the filename for update: ");

	scanf("%s", filename);

	if(stat(filename, &buf) < 0) {
		fprintf(stderr, "Unknown file state\n");
		return  -1;		
	}
	mode = buf.st_mode;
	if (! S_ISREG(mode)) {
		fprintf(stderr, "Unknown file type\n");
		return -1;
	}

	if(getFileSize(filename) != FLASH_SIZE) {
		fprintf(stderr, "Error: The size of file %s is not correct, please check it! \n", filename);
		return	-2;
	}
	
	if ((fp = fopen(filename, "rb")) == NULL)
	{
		fprintf(stderr, "Cannot open file: %s\n", filename);
		return  -1;
	}	
	
	struct sigaction act;
	//-act.sa_handler=ctrl_c_op;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "Error\n");
		return -1;
	}

	fseek(fp, 0, SEEK_SET);
	for (i = 0; i < FLASH_SIZE; i++)
	{
		rpd = fgetc(fp);
		source[i] = exchange(rpd);
	}

	vendorid = source[0x7FFE0];
	vendorid = (vendorid << 8) + source[0x7FFE1];
	devid =  source[0x7FFE2];
	devid = (devid << 8) + source[0x7FFE3];
	
	if( (devid != cur_pci_dev->device_id) || (vendorid != cur_pci_dev->vendor_id) ) {		
		while(1) {
		    fprintf(stderr, "warning: The input file %s id %04x:%04x is not matching with the card id %04x:%04x,\n", 
		        filename, vendorid,devid,cur_pci_dev->vendor_id,cur_pci_dev->device_id);
		    fprintf(stderr, "warning: continue to use this file may cause the board failure and cannot be recovered,\n");
		    fprintf(stderr, "warning: you can type \"U\" to use it anyway, or any others to quit this program. \n\n");
			setbuf(stdin, NULL);
			memset(char_input, 0, 4);			
			i = 0;
			while((ch = getchar()) != '\n') {
				char_input[i++] = ch;
				if (i == 4) {
					break;
				}
			}

			if (char_input[0] == '\0') {
				continue;
			}
			
			if (char_input[0] == 'U' || char_input[0] == 'u') {
				break;
			} else {
				return	-2;
			}
		}				
	}
	//printf("\t---------------Debug: Get File - VendorID:0x%x  DeviceID:0x%x\n", vendorid, devid);

	//flashȫ
	printf("Erasing epcs flash...\n");
	syslog(LOG_INFO, "Erasing epcs flash...\n");
	alt_epcs_flash_erase_bulk(dev);
	printf("Erasing epcs flash done\n");
	syslog(LOG_INFO, "Erasing epcs flash done\n");

	//ýȰٷֱ
	printf("Writing epcs flash...\n");
	syslog(LOG_INFO, "Writing epcs flash...\n");

	ret_code = alt_epcs_flash_write_block(dev, 0, source, FLASH_SIZE);
	if (ret_code)
	{
		fprintf(stderr, "Writng epcs flash error.\n");
		fclose(fp);
		return -2;
	}

	fclose(fp);

	act.sa_handler=NULL;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "Error\n");
		return -1;
	}

	printf("Writing epcs flash successfully\n");
	syslog(LOG_INFO, "Writing epcs flash successfully\n");

	return 0;
}


static int epcs_flash_all_update_rpd(t_opvx_pci_dev_s *cur_pci_dev)
{
	char filename[32];
	unsigned char rpd;
	int ret_code;
	unsigned char source[FLASH_SIZE];
	int i,j;
	FILE *fp = NULL;
	unsigned char ch;
	struct alt_flash_epcs_dev *epcs;
	struct alt_flash_dev *dev;
	struct sigaction act;
	unsigned char dest[EPCS_CONFIG_PAGE_SIZE];
	struct stat buf;
	mode_t mode;
		
	//--------------------------------------------------------------------------
	if (!popvx_pci_dev) {
		fprintf(stderr, "No OpenVox cards found!\n");
		return -1;
	}
	if (!cur_pci_dev || !cur_pci_dev->epcs_flash_dev) {
		fprintf(stderr, "No EPCS controller found.\n");	
		return -1;
	}

	epcs = cur_pci_dev->epcs_flash_dev;

	//--------------------------------------------------------------------------
	// ð忨
	int card_type = cur_pci_dev->card_type;

	//--------------------------------------------------------------------------
	// pin change flag
	int pin_chg = cur_pci_dev->pin_chg;

#ifdef DEBUG
	printf(DEBUGFMT "Card Name = %s Card Type = %d\n", DEBUGARGS, cur_pci_dev->card_name, cur_pci_dev->card_type);
#endif
	//--------------------------------------------------------------------------
	// get interrupt pin code
	// 0xFF: ʾıinterrupt pin
	ch = epcs_modify_interrupt_pin(pin_chg);

	//--------------------------------------------------------------------------
	// ޸interrupt pin
	if (0xFF != ch) {
		for(i = 0; i < max_epcs_dev; i ++) {
			if (card_indexes[i] != card_type) {
				continue;
			}
			
			cur_pci_dev = popvx_pci_dev + i;
			epcs = cur_pci_dev->epcs_flash_dev;
			if (!epcs) {
				fprintf(stderr, "No EPCS controller found.\n");	
				continue;
			}
			dev	= &(epcs->dev);
			if (!dev) {
				fprintf(stderr, "No EPCS flash found.\n");	
				continue;
			}

			// ӳepcsĴ
			if (epcs_map_reg(cur_pci_dev) != 0) {
				fclose(fp);
				return -2;
			}
			
			// ʼepcs flash
			if (alt_epcs_flash_init(epcs) < 0)
			{
				fprintf(stderr, "Initializing flash error.\n");
				epcs_unmap_reg(cur_pci_dev);
				if (fp) {
					fclose(fp);
				}
				return -2;
			}

			//-act.sa_handler=ctrl_c_op;
			act.sa_flags=0;
			if(sigaction(SIGINT, &act, NULL) < 0)
			{
				fprintf(stderr, "error\n");
				return -1;
			}

			// ȡöϢ
			printf("Reading epcs flash...\t\t\t\t[%d]\n", i + 1);
			syslog(LOG_INFO, "Reading epcs flash...\n");
			ret_code = alt_epcs_flash_read(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
			if (ret_code)
			{
				fprintf(stderr, "Reading epcs flash error.\n");
				return -2;
			}
			printf("Reading epcs flash done\t\t\t\t[%d]\n", i + 1);
			syslog(LOG_INFO, "Reading epcs flash done\n");


			// ö
			printf("Erasing epcs flash...\t\t\t\t[%d]\n", i + 1);
			syslog(LOG_INFO, "Erasing epcs flash...\n");
			alt_epcs_flash_erase_block(dev, EPCS_CONFIG_PAGE_OFFSET);
			printf("Erasing epcs flash done\t\t\t\t[%d]\n", i + 1);
			syslog(LOG_INFO, "Erasing epcs flash done\n");

			// дµö
			printf("Writing epcs flash...\t\t\t\t[%d]\n", i + 1);
			syslog(LOG_INFO, "Writing epcs flash...\n");

			// change interrupt pin register value
			dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_INTS] = ch ;
			
			ret_code = alt_epcs_flash_write_block(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
			if (ret_code)
			{
				fprintf(stderr, "Writng epcs flash error.\n");
				return -2;
			}

			epcs_unmap_reg(cur_pci_dev);

			act.sa_handler=NULL;
			act.sa_flags=0;

			if(sigaction(SIGINT, &act, NULL) < 0)
			{
				fprintf(stderr, "Error\n");
				return -1;
			}
			
			printf("Writing epcs flash successfully\t\t\t[%d]\n", i+1);
			syslog(LOG_INFO, "Writing epcs flash successfully\n");

			C3Reload_run(cur_pci_dev);

			sleep(2);
			
		}
					
	//--------------------------------------------------------------------------
	// ļи¹̼
	} else {
		printf("Please input the filename for update: ");
		scanf("%s", filename);

		if(stat(filename, &buf) < 0) {
			fprintf(stderr, "Unknown file state\n");
			return  -1;		
		}
		mode = buf.st_mode;
		if (! S_ISREG(mode)) {
			fprintf(stderr, "Unknown file type\n");
			return -1;
		}
		if ((fp = fopen(filename, "rb")) == NULL)
		{
			fprintf(stderr, "Cannot open file: %s\n", filename);
			return  -1;
		}	

		for(i = 0; i < max_epcs_dev; i ++) {
			if (card_indexes[i] != card_type) {
				continue;
			}
			
			cur_pci_dev = popvx_pci_dev + i;
			epcs = cur_pci_dev->epcs_flash_dev;
			if (!epcs) {
				fprintf(stderr, "No EPCS controller found.\n");	
				continue;
			}
			
			dev	= &(epcs->dev);
			if (!dev) {
				fprintf(stderr, "No EPCS flash found.\n");	
				continue;
			}
			
			//-act.sa_handler=ctrl_c_op;
			act.sa_flags=0;
			if(sigaction(SIGINT, &act, NULL) < 0)
			{
				fprintf(stderr, "Error\n");
				if (fp) {
					fclose(fp);
				}
				return -1;
			}
			
			// ӳepcsĴ
			if (epcs_map_reg(cur_pci_dev) != 0) {
				fclose(fp);
				return -2;
			}
			
			// ʼepcs flash
			if (alt_epcs_flash_init(cur_pci_dev->epcs_flash_dev) < 0)
			{
				fprintf(stderr, "Initializing flash error.\n");
				epcs_unmap_reg(cur_pci_dev);
				if (fp) {
					fclose(fp);
				}
				return -2;
			}			
		
			//flashȫ
			printf("Erasing epcs Flash...\t\t\t\t[%d]\n", i+1);
			syslog(LOG_INFO, "Erasing epcs Flash...\n");
			alt_epcs_flash_erase_bulk(&(epcs->dev));
			printf("Erasing epcs flash done\t\t\t\t[%d]\n", i+1);
			syslog(LOG_INFO, "Erasing epcs flash done\n");

			//ýȰٷֱ
			printf("Writing epcs flash...\t\t\t\t[%d]\n", i+1);
			syslog(LOG_INFO, "Writing epcs flash...\n");

			fseek(fp, 0, SEEK_SET);
			for (j = 0; j < FLASH_SIZE; j++)
			{
				rpd = fgetc(fp);
				source[j] = exchange(rpd);
			}

			ret_code = alt_epcs_flash_write_block(dev, 0, source, FLASH_SIZE);
			if (ret_code)
			{
				fprintf(stderr, "Writng epcs flash error.\n");
				if (fp) {
					fclose(fp);
				}
				return -2;
			}
			
			epcs_unmap_reg(cur_pci_dev);

			act.sa_handler=NULL;
			act.sa_flags=0;

			if(sigaction(SIGINT, &act, NULL) < 0)
			{
				fprintf(stderr, "Error\n");
				if (fp) {
					fclose(fp);
				}
				return -1;
			}
			
			printf("Writing epcs flash successfully\t\t\t[%d]\n", i+1);
			syslog(LOG_INFO, "Writing epcs flash successfully\n");

			C3Reload_run(cur_pci_dev);

			sleep(2);
		}
			
		if (fp) {
			fclose(fp);
		}
	}


	
	return 0;
}
static int epcs_flash_edit_rpd(t_opvx_pci_dev_s *cur_pci_dev)
{
		int ret_code, i;
		alt_u8	siliconid;
		//alt_u8  hexvalue;
		unsigned char dest[EPCS_CONFIG_PAGE_SIZE];
		unsigned short int *pConfigPage=NULL,*pVendorID=NULL,*pDeviceID=NULL;

	if (!cur_pci_dev || !cur_pci_dev->epcs_flash_dev) {
		fprintf(stderr, "No EPCS controller found.\n");	
		return -1;
	}
	struct alt_flash_epcs_dev *epcs = cur_pci_dev->epcs_flash_dev;
	struct alt_flash_dev *dev		= &(epcs->dev);
	
	struct sigaction act;
	//-act.sa_handler=ctrl_c_op;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}

	printf("Reading epcs flash...\n");

	//read config page from flash epcs
	ret_code = alt_epcs_flash_read(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
	
	if (ret_code)
	{
		fprintf(stderr, "Reading epcs flash error.\n");
		return -1;
	}

	act.sa_handler=NULL;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}
		//show the read data in hex matrix
		//-int	i;
		//-for (i=0; i<EPCS_CONFIG_PAGE_SIZE; i++)
		//-  {
		//-	if (! (i & 15))
		//-  printf("%02x:", i);
		//-	printf(" %02x", dest[i]);
		//-	if ((i & 15) == 15)
		//- putchar('\n');
		//-}

	printf("Reading epcs flash successfully\n");
	//=====================!!!!!! Read and show the config page data !!!!!!!================================
	/* Send the RES command sequence */
	// get EPCS silicon ID
	  siliconid = epcs_read_electronic_signature(cur_pci_dev->epcs_flash_dev->register_base);
	  printf("siliconID = 0x%02x \n", siliconid);


	//=====================!!!!!! Show the option to choose and prepare data for writing !!!!!!!============
		pVendorID=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_VENDORID_H];
		pDeviceID=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_DEVICEID_H];
		//printf("Current VendorID=0x%x DeviceID=0x%x \n",*pVendorID, *pDeviceID);

		printf("choose Vendor to modify: 1. Digium(0xb161)  2.OpenVox(0x1b74)\n");
		printf("[Press 1-2 to modify, or other key to quit] : ");
		unsigned char ch,buf[4];
		int keyret;

	while(1)
	{
		
		setbuf(stdin, NULL);
		memset(buf, 0, 4);
		
		i = 0;
		while((ch = getchar()) != '\n') {
			buf[i++] = ch;
			if (i == 4) {
				break;
			}
		}

		if (buf[0] == '\0') {
			fprintf(stderr, "\nError: Invalid option, input an integer\n");
			continue;
		}
		
		buf[3] = '\0';
		keyret = atoi(buf);
		//-printf("buf = %s\n", buf);
		//-printf("keyret = %d\n", keyret);
		break;
	}	

		switch(keyret) {
				case 1:
					*pVendorID=0x61d1;
					if(*pDeviceID==0x30d4)
					*pDeviceID=0x2014;
					else if(*pDeviceID==0x30d2)
					*pDeviceID=0x2012;
					else if(*pDeviceID==0x30d1)
					{
					printf("No Digium products type to match! \n");
					return 0;
					}
					else if(*pDeviceID==0x10d4)
					*pDeviceID=0x1014;
					else if(*pDeviceID==0x10d2)
					*pDeviceID=0x1012;
					else if(*pDeviceID==0x10d1)
					{
					printf("No Digium products type to match! \n");
					return 0;
					}
					pConfigPage=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_VENDORID_H];
					*pConfigPage=*pVendorID;
					//printf("wujc-debug: VendorID=0x%x...\n", *pConfigPage);
					pConfigPage=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_DEVICEID_H];
					*pConfigPage=*pDeviceID;
					//printf("wujc-debug: DeviceID=0x%x...\n", *pConfigPage);
					break;
				case 2:
					*pVendorID=0x741b;
					if(*pDeviceID==0x2014)
					*pDeviceID=0x30d4;
					else if(*pDeviceID==0x2012)
					*pDeviceID=0x30d2;
					else if(*pDeviceID==0x1014)
					*pDeviceID=0x10d4;
					else if(*pDeviceID==0x1012)
					*pDeviceID=0x10d2;
					pConfigPage=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_VENDORID_H];
					*pConfigPage=*pVendorID;
					//printf("wujc-debug: VendorID=0x%x...\n", *pConfigPage);
					pConfigPage=&dest[EPCS_CONFIG_RES_OFFSET + EPCS_CONFIG_REG_OFFSET_DEVICEID_H];
					*pConfigPage=*pDeviceID;
					//printf("wujc-debug: DeviceID=0x%x...\n", *pConfigPage);
					break;
				default:
					return 0;
					break;
		}

	//Prompt input 
	printf("Please input the serial number(0x00-0xFF): 0x");
	while(1)
	{
		
		setbuf(stdin, NULL);
		setbuf(stdout, NULL);
		fflush(stdin);
		fflush(stdout);
		memset(buf, 0, 4);
		
		i = 0;
		while((ch = getchar()) != '\n') {
			buf[i++] = ch;
			//printf("input=%d\n",ch);
			if (i > 32) {
				break;
			}
		}

		if (buf[0] == '\0'||buf[2] != '\0'||buf[1] =='\0') {
			fprintf(stderr, "\nError: Invalid option, input an integer\n");
			continue;
		}
		//-buf[3] = '\0';
		keyret = atoi(&buf[0]);
		//printf("buf = %s\n", buf);
		//printf("keyret = %d\n", keyret);
		break;
	}
	//convert char to hex value 
	ch=buf[0];
	if(ch>47 && ch<58) ch -= 48;
	else if(ch>64 && ch<71) ch-=55;
	else if(ch>96 && ch<103) ch-=87;
	//buf[0]=char2hex(buf[0]);

	if(buf[1]>47 && buf[1]<58) buf[1]-=48;
	else if(buf[1]>64 && buf[1]<71) buf[1] -=55;
	else if(buf[1]>96 && buf[1]<103) buf[1]-=87;
	//buf[1]=char2hex(buf[1]);

	//hexvalue=char2hex(buf[0])*16+char2hex(buf[1]);
	dest[EPCS_CONFIG_RES_OFFSET+EPCS_CONFIG_REG_OFFSET_SUBVENDORID_L]=buf[0]*16+buf[1];
	//printf("wujc-debug: Revision=0x%x\n",dest[EPCS_CONFIG_RES_OFFSET+EPCS_CONFIG_REG_OFFSET_SUBVENDORID_L]);


		//===============================================================================
		//for (i=0; i<EPCS_CONFIG_PAGE_SIZE; i++)
		//  {
		//	if (! (i & 15))
		//  printf("%02x:", i);
		//	printf(" %02x", dest[i]);
		//	if ((i & 15) == 15)
		//  putchar('\n');
		//  }
		//===============================================================================

	//-if(keyret>=00 && keyret<=255)
	//-{
	//-	printf("okay judge correct!\n");
	//-	return 0;
	//-}
	//=====================!!!!!! Below do erase the page and read again !!!!!!!================================
	//-act.sa_handler=ctrl_c_op;
	act.sa_flags=0;

	if(sigaction(SIGINT, &act, NULL) < 0)
	{
		fprintf(stderr, "error\n");
		return -1;
	}
			// ӳepcsĴ
			//-wujcif (epcs_map_reg(cur_pci_dev) != 0) {
			//-wujc	return -2;
			//-wujc}

		// do page erase operation 
		printf("Erasing epcs flash...\t\t\t\t\n");
		alt_epcs_flash_erase_block(dev, EPCS_CONFIG_PAGE_OFFSET);
		printf("Erasing epcs flash done\t\t\t\t\n");

		ret_code = alt_epcs_flash_write_block(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
		if (ret_code)
		{
			fprintf(stderr, "Writng epcs flash error.\n");
			//-free(epcs);
			return -2;
		}
		printf("Writing configuration data to epcs flash successfully\t\t\t\n");

	ret_code = alt_epcs_flash_read(dev, EPCS_CONFIG_PAGE_OFFSET, dest, EPCS_CONFIG_PAGE_SIZE);
	
	if (ret_code)
	{
		fprintf(stderr, "Reading epcs flash error.\n");
		return -1;
	}
		//===============================================================================
		//for (i=0; i<EPCS_CONFIG_PAGE_SIZE; i++)
		//  {
		//	if (! (i & 15))
		//  printf("%02x:", i);
		//	printf(" %02x", dest[i]);
		//	if ((i & 15) == 15)
		//  putchar('\n');
		//  }
		//===============================================================================

	/* Send the RES command sequence */
	// get EPCS silicon ID
	siliconid = epcs_read_electronic_signature(cur_pci_dev->epcs_flash_dev->register_base);
	  printf("siliconID = %02x \n", siliconid);

	  printf("Please reboot system to take effect!\n");
			//-wujc epcs_unmap_reg(cur_pci_dev);

			act.sa_handler=NULL;
			act.sa_flags=0;

			if(sigaction(SIGINT, &act, NULL) < 0)
			{
				fprintf(stderr, "Error\n");
				return -1;
			}

		return 0;
}
#if 0
alt_u8 char2hex (unsigned char ichar)
{
	//unsigned short int 
	alt_u8 ch;
	ch=(alt_u8)ichar;
	if(ch>47 && ch<58)	ch -= 48;
	else if(ch>64 && ch<71)	ch-=55;
	else if(ch>96 && ch<103)	ch-=87;
	else	{
		printf("invalid input !");
		return -1;
		}
	return ch;
}
#endif
static int lockfile(int fd)
{
	struct flock fk;
	fk.l_type 	= F_WRLCK;
	fk.l_start	= 0;
	fk.l_whence	= SEEK_SET;
	fk.l_len	= 0;
	
	return (fcntl(fd, F_SETLK, &fk));
}

#define EPCS_INFO \
"########################################################################\n" \
"#                   OpenVox EPCS Flash Update Utility                  #\n" \
"#                            %s                            #\n"     \
"#                     OpenVox Communication Co.,Ltd                    #\n" \
"#        Copyright (c) 2009-2011 OpenVox. All Rights Reserved.         #\n" \
"########################################################################\n" \
"Note: make sure the driver is not runing before update firmware.\n"      \
"      Do not turn power OFF while the update process is continuing.\n"      \
"      When the update is completed, the system will reboot.\n\n"

static void get_info(void) {
	printf(EPCS_INFO, version);
}

/************************************************************************
 * Main Function
 ************************************************************************/
int main(void)
{	
	int ret = 0;
	int ret_tmp = 0;
	int err_flag = 0;
	int menu_status = MENU_STATUS_1;
	int cur_epcs_flash_idx;
	t_opvx_pci_dev_s * cur_pci_dev = NULL;
	enum enum_menu_card_item menu_oper_item = MENU_CARD_OPER_NONE;

	int fd = 0;
	char buf[6];
	
	/* Check if we're root */	
	if (geteuid()) {
		fprintf(stderr, "Must be run as root\n");
		exit(1);
	}
	
	/* Singleton pattern */
	fd = open(EPCS_PID, O_RDWR|O_CREAT, S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
	if (fd < 0) {
		fprintf(stderr, "Unable to open pid file '%s': %s\n", (char *)EPCS_PID, strerror(errno));
		exit(1);
	}
	if(lockfile(fd) < 0) {
		if(errno == EACCES  || errno == EAGAIN) {
			fprintf(stderr, "EPCS already running on '%s'\n", (char *)EPCS_PID);
			close(fd);
			return 1;
		}
		fprintf(stderr, "Can't lock %s : %s", EPCS_PID, strerror(errno));
		exit(1);
	} 

	ftruncate(fd, 0);
	sprintf(buf, "%ld\n", (long int)getpid());
	write(fd, buf, strlen(buf) + 1);
//	close(fd);
	
	get_info();
	
	epcs_flash_init();
	
	while (1) {
		switch(menu_status) {

			case MENU_STATUS_1:
				// scan openvox card's by vendor_id and device_id
				menu_oper_item = menu_choose(menu_main, num_of_menu_items(menu_main));
				if (0 == epcs_flash_scan(menu_oper_item))
				{
					fprintf(stderr, "No OpenVox cards found.\n");
					return  -ENODEV;
				}

				menu_status = MENU_STATUS_2;
				break;
				
			case MENU_STATUS_2:
				ret  = menu_choose(pmenu_epcs_flash, (MENU_CARD_UPDATE == menu_oper_item) ? max_epcs_dev + card_type_number : max_epcs_dev);
				if ((ret >= 1) && (ret <= max_epcs_dev)) {
					cur_epcs_flash_idx = ret - 1;
					if (cur_epcs_flash_idx < 0 && cur_epcs_flash_idx >= max_epcs_dev) {
						fprintf(stderr, "OpenVox card id error.\n");
						err_flag = 1;
						break;
					}
					cur_pci_dev = &popvx_pci_dev[cur_epcs_flash_idx];
					if (!cur_pci_dev) {
						fprintf(stderr, "No OpenVox PCI Device.\n");
						err_flag = 1;
						break;
					}
					// mmap 
					if (epcs_map_reg(cur_pci_dev) != 0) {
						err_flag = 1;
						break;
					}

					// init epcs flash
					if (alt_epcs_flash_init(cur_pci_dev->epcs_flash_dev) < 0){
						fprintf(stderr, "Initializing flash error.\n");
						err_flag = 1;
						break;
					}
					err_flag = 0; // ok
					menu_status = MENU_STATUS_3;

				}
				else if ((max_epcs_dev + 1 <= ret) && (ret <= max_epcs_dev + card_type_number) && (MENU_CARD_UPDATE == menu_oper_item)) {
					err_flag = 0; // ok
					menu_status = MENU_STATUS_3;
				} else if (0 == ret) {
					epcs_flash_destroy();
					return 0;
				} else {
					continue;	
				}
				break;
				
			case MENU_STATUS_3:
//				ret  = menu_choose(menu_flash_operation, num_of_menu_items(menu_flash_operation));
				if ((ret >= 1) && (ret <= max_epcs_dev)) {
					//----------------------------------------------------------
					// ¼flash
					if (MENU_CARD_UPDATE == menu_oper_item) {
						ret_tmp = epcs_flash_update_rpd(cur_pci_dev);
						if (ret_tmp == 0) {
							C3Reload_run(cur_pci_dev);						
						} else if (ret_tmp == -1) {
							continue;
						} else {
							err_flag = 1;
						}
					//----------------------------------------------------------
					// ȡflash	
					} else if (MENU_CARD_BACKUP == menu_oper_item){
						ret_tmp = epcs_flash_backup_rpd(cur_pci_dev);
						menu_status = MENU_STATUS_2;
						continue;
					}
					// Add entry edit the flash by w01 -s
					else if (MENU_CARD_EDIT == menu_oper_item) {
						ret_tmp = epcs_flash_edit_rpd(cur_pci_dev);
						if (0 == ret_tmp) {
							epcs_flash_destroy();
							return 0;
						}
						menu_status = MENU_STATUS_2;
						continue;
						}
					// w01 -e
				}  else if ((max_epcs_dev + 1 <= ret) && (ret <= max_epcs_dev + card_type_number) && (MENU_CARD_UPDATE == menu_oper_item)) {
					cur_pci_dev = &popvx_pci_dev[pmenu_epcs_flash[ret - 1].device_idx];
#ifdef DEBUG
					printf("pmenu_epcs_flash[%d]\n", ret);
#endif
					ret_tmp = epcs_flash_all_update_rpd(cur_pci_dev);
					if (ret_tmp == -1) {
						continue;
					} else if (ret_tmp == -2) {
						err_flag = 1;
					}
				}		

				if (0 == err_flag && MENU_CARD_UPDATE == menu_oper_item) {
					printf("\nPress ENTER key to reboot your system.\n");	
					setbuf(stdin, NULL);
					getchar();
					system("reboot");			
					exit(0);			
				}
				break;
				
			default:
				break;
		}// switch end

		if (err_flag == 1) {
			if ((ret >= 1) && (ret <= max_epcs_dev + card_type_number)) {
				// unmap
				if (ret != max_epcs_dev + 1) {
					epcs_unmap_reg(cur_pci_dev);					
				}
				// close pid file
				if (fd) {
					close(fd);
				}
				epcs_flash_destroy();
			}
			exit(-1);
		} 
	}
	

	exit(0);
}

