/***********************************************************
* Copyright (c) 2009-2010 OpenVox Communication Co.,Ltd.
* All Rights Reserved.	
* Author: mark.liu@openvox.cn
***********************************************************/
/*
 *	A flash update programme for OpenVox G400P card
 */

#ifndef _EPCS_H_
#define _EPCS_H_

#include "alt_types.h"

#define DBGHEAD __FILE__ ":%d %s: "
#define DBGINFO __LINE__,__PRETTY_FUNCTION__

#define DEBUGFMT  "[%s(%d)-%s]  "
#define DEBUGARGS __FILE__,__LINE__,__FUNCTION__

#define EPCS_PID	"/var/run/epcs.pid"
/*
 * epcs_controller0 configuration
 */
#define EPCS_CONTROLLER0_NAME "/dev/epcs_controller0"
#define EPCS_CONTROLLER0_TYPE "altera_avalon_epcs_flash_controller"
#define EPCS_CONTROLLER0_BASE 0x00002800
#define EPCS_CONTROLLER0_SPAN 2048
#define EPCS_CONTROLLER0_IRQ ALT_IRQ_NOT_CONNECTED
#define EPCS_CONTROLLER0_DATABITS 8
#define EPCS_CONTROLLER0_TARGETCLOCK 20
#define EPCS_CONTROLLER0_CLOCKUNITS "MHz"
#define EPCS_CONTROLLER0_CLOCKMULT 1000000
#define EPCS_CONTROLLER0_NUMSLAVES 1
#define EPCS_CONTROLLER0_ISMASTER 1
#define EPCS_CONTROLLER0_CLOCKPOLARITY 0
#define EPCS_CONTROLLER0_CLOCKPHASE 0
#define EPCS_CONTROLLER0_LSBFIRST 0
#define EPCS_CONTROLLER0_EXTRADELAY 0
#define EPCS_CONTROLLER0_TARGETSSDELAY 100
#define EPCS_CONTROLLER0_DELAYUNITS "us"
#define EPCS_CONTROLLER0_DELAYMULT "1e-006"
#define EPCS_CONTROLLER0_PREFIX "epcs_"
#define EPCS_CONTROLLER0_REGISTER_OFFSET 0x400
#define EPCS_CONTROLLER0_USE_ASMI_ATOM 0
#define EPCS_CONTROLLER0_CLOCKUNIT "kHz"
#define EPCS_CONTROLLER0_DELAYUNIT "us"

#define EPCS_CONTROLLER0_REGISTER_OFFSET 0x400

#define ALT_MODULE_CLASS_epcs_controller0 altera_avalon_epcs_flash_controller


//#include "epcs_commands.h"
#define epcs_read    0x03
#define epcs_pp      0x02
#define epcs_wren    0x06
#define epcs_wrdi    0x04 //
#define epcs_rdsr    0x05
#define epcs_wrsr    0x01
#define epcs_se      0xD8
#define epcs_be      0xC7
#define epcs_dp      0xB9 //
#define epcs_res     0xAB

//#include "errno.h"
#define EIO 5		/* I/O error */
#define ENOMEM 12	/* Not enough core */
#define ENODEV 19	/* No such device */
#define EINVAL 22	/* Invalid argument */

/*******************************************************************************
 * EPCS4 :
	Bytes (bits) 				524,888 bytes (4 Mbits)
	Number of sectors 			8
	Bytes (bits) per sector 	65,536 bytes (512 Kbits)
	Pages per sector 			256
	Total number of pages 		2,048
	Bytes per page 				256 bytes
 ******************************************************************************/
 /*******************************************************************************
 * EPCS4 Sector Address:
	Sector			Address Range (Byte Addresses in HEX)
					Start 		End
	7 				H'70000 	H'7FFFF  for configuration
	6 				H'60000 	H'6FFFF
	5 				H'50000	 	H'5FFFF
	4 				H'40000	 	H'4FFFF
	3 				H'30000	 	H'3FFFF
	2 				H'20000	 	H'2FFFF
	1 				H'10000 	H'1FFFF
	0 				H'00000 	H'0FFFF
 ******************************************************************************/
// Config block info
//#define EPCS_CONFIG_SECTOR_OFFSET 0x70000
#define EPCS_CONFIG_PAGE_OFFSET  0x07ff00
#define EPCS_CONFIG_PAGE_SIZE    256

#define EPCS_CONFIG_RES_OFFSET   224
#define EPCS_CONFIG_REG_OFFSET_VENDORID_H  0
#define EPCS_CONFIG_REG_OFFSET_VENDORID_L  1
#define EPCS_CONFIG_REG_OFFSET_DEVICEID_H  2
#define EPCS_CONFIG_REG_OFFSET_DEVICEID_L  3

#define EPCS_CONFIG_REG_OFFSET_REVISION  7

#define EPCS_CONFIG_REG_OFFSET_SUBDEVICEID_H  8
#define EPCS_CONFIG_REG_OFFSET_SUBDEVICEID_L  9
#define EPCS_CONFIG_REG_OFFSET_SUBVENDORID_H  10
#define EPCS_CONFIG_REG_OFFSET_SUBVENDORID_L  11

#define EPCS_CONFIG_REG_OFFSET_INTS  15

// card type
enum enum_card_type{
	OPVX_NONE,
	OPVX_G400P,
	OPVX_A2410P,
	OPVX_A1610,
	OPVX_A810,
	WC_D830P,
	WC_D430P,
	OPVX_D430P,
	WC_D230P,
	OPVX_D230P,
	OPVX_D130P,	
	WC_D410G2,
	OPVX_D410G2,
	WC_D210G2,
	OPVX_D210G2,
	OPVX_D115G2,
	OPVX_X200,
	OPVX_END
};

// menu card itmes
enum enum_menu_card_item{
	MENU_CARD_NONE,
	MENU_CARD_G400P,
	MENU_CARD_A2410P,
	MENU_CARD_A1610,
	MENU_CARD_A810,
	MENU_CARD_WC_D430P,
	MENU_CARD_D430P,
	MENU_CARD_WC_D230P,
	MENU_CARD_D230P,
	MENU_CARD_D130P,
	MENU_CARD_WC_D410G2,
	MENU_CARD_D410G2,
	MENU_CARD_WC_D210G2,
	MENU_CARD_D210G2,
	MENU_CARD_D115G2,
	MENU_CARD_X200,
	MENU_CARD_END
};

enum enum_menu_oper_item{
	MENU_CARD_OPER_NONE,
	MENU_CARD_UPDATE,
	MENU_CARD_BACKUP,
	MENU_CARD_EDIT
};

//#include "alt_errno.h"
//#include "altera_avalon_epcs_flash_controller.h"
/*
 * alt_llist is the structure used to represent an element within a linked
 * list.
 */
typedef struct alt_llist_s alt_llist;

struct alt_llist_s
{
    alt_llist *next; /* Pointer to the next element in the list. */
    alt_llist *previous; /* Pointer to the previous element in the list. */
};

/*
 * ALT_LLIST_HEAD is a macro that can be used to create the head of a new
 * linked list. This is named "head". The head element is initialised to
 * represent an empty list.
 */

#define ALT_LLIST_HEAD(head) alt_llist head = {&head, &head}

/*
 * ALT_LLIST_ENTRY is a macro used to define an uninitialised linked list
 * entry. This is used to reserve space in structure initialisation for
 * structures that inherit form alt_llist.
 */

#define ALT_LLIST_ENTRY {0, 0}

/*
 * The alt_dev_llist is an internal structure used to form a common base
 * class for all device types. The use of this structure allows common code
 * to be used to manipulate the various device lists.
 */
typedef struct
{
    alt_llist llist;
    const char *name;
} alt_dev_llist;

/*
 * Description of a single Erase region
 */
typedef struct flash_region
{
    int offset;
    int region_size;
    int number_of_blocks;
    int block_size;
} flash_region;

typedef struct alt_flash_dev alt_flash_dev;
typedef alt_flash_dev alt_flash_fd;

typedef alt_flash_dev *(*alt_flash_open)(alt_flash_dev *flash, const char *name);
typedef int(*alt_flash_close)(alt_flash_dev *flash_info);

typedef int(*alt_flash_write)(alt_flash_dev *flash, int offset, const void *src_addr, int length);

typedef int(*alt_flash_get_flash_info)(alt_flash_dev *flash, flash_region **info, int *number_of_regions);
typedef int(*alt_flash_write_block)(alt_flash_dev *flash,  //int block_offset,
int data_offset, const void *data, int length);
typedef int(*alt_flash_erase_block)(alt_flash_dev *flash, int offset);
typedef int(*alt_flash_read)(alt_flash_dev *flash, int offset, void *dest_addr, int length);

#define ALT_MAX_NUMBER_OF_FLASH_REGIONS 8

struct alt_flash_dev
{
    alt_llist llist;
    const char *name;
    alt_flash_open open;
    alt_flash_close close;
    alt_flash_write write;
    alt_flash_read read;
    alt_flash_get_flash_info get_info;
    alt_flash_erase_block erase_block;
    alt_flash_write_block write_block;
    unsigned char *base_addr;
    int length;
    int number_of_regions;
    flash_region region_info[ALT_MAX_NUMBER_OF_FLASH_REGIONS];
};

typedef struct alt_flash_epcs_dev t_alt_flash_epcs_dev;

/*
 *  Description of the flash
 *
 *  Contains the basic alt_flash_dev, plus
 *  epcs-specific parameters.
 *
 *  Every parameter that distinguishes this
 *  serial flash device from any other is
 *  encoded here.
 *
 *  Example:
 *  size_in_bytes: the number of bytes of
 *  storage in this chip.
 */
struct alt_flash_epcs_dev
{
    alt_flash_dev dev;

    //-alt_u32 register_base;
    unsigned char *register_base;
    alt_u32 size_in_bytes;
    alt_u32 silicon_id;
    alt_u32 page_size;
};

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

#define FLASH_SIZE 512*1024 // 512Kb
///////////////////////////////////////////////////////////
// epcs flash definition

#define EPCS_MAX_DEVICE 128

enum epcs_flash_operation
{
	EPCS_FLASH_OPERATION_NONE,
	EPCS_FLASH_OPERATION_BACKUP,
	EPCS_FLASH_OPERATION_UPDATE,
	EPCS_FLASH_OPERATION_EXIT
};

enum enum_pin
{
	PIN_CHANGED_DISABLE,
	PIN_CHANGED_ENABLE
};

struct opvx_pci_dev_s
{
	struct alt_flash_epcs_dev *epcs_flash_dev;
	/* epcs controller0 */
	unsigned int id;
	//-unsigned int base;
	unsigned char *base;
	unsigned char *epcs_reg;	
	unsigned int baseoff;
	unsigned int mem_base;
	unsigned int register_offset;

	/* C3Reload*/
	unsigned char *c3reload_reg;	
	//-unsigned int c3reload_base;
	unsigned char *c3reload_base;
	unsigned int c3reload_baseoff;
	unsigned int c3reload_register_offset;	
	
	char *manufacturer;
	char *card_name;

	unsigned int card_type; 

	/* struct pci_dev */
	u16 domain;				/* PCI domain (host bridge) */
	u8 bus, dev, func;			/* Bus inside domain, device and function */
	u16 vendor_id, device_id;		/* Identity of the device */
	int irq;				/* IRQ number */

	/* Interrupt Pin */
	u8 pin;

	enum enum_pin pin_chg;	
};
typedef struct opvx_pci_dev_s t_opvx_pci_dev_s;



#endif // _EPCS_H_

